package rl.simplerl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import rl.AbstractReinforcementLearning;
import rule.ActiveComboRuleWrapper;
import rule.EmptyRule;
import rule.GameStateInterface;
import rule.Rule;

/**
 * Simple Reinforcement Learning Algorithm that generates scripts of variable
 * length from a static rulebase.
 * 
 * @author Kevin Majchrzak
 * 
 */
public class SimpleReinforcementLearning extends AbstractReinforcementLearning {

	private static final long UPDATE_FREQUENCY_MILLIS = 4000;
	private static final float BREAK_EVEN = 0.8f;
	private static final long W_MAX_HIGH = 300;
	private static final long W_MIN = 0;
	private static final long P_MAX = 80;
	private static final long R_MAX = 80;
	private static final long INITIAL_WEIGHT = 80;
	private static final long W_MAX_LOW = 200;
	private RuleWrapper[] script;
	private boolean updateScript;

	private RuleWrapper[] ruleWrapperBase;
	private long lastUpdateSystemMillis;
	private int enemyOldHP, myOldHP;
	private long sumWeights;
	private RuleWrapper[] defaultRuleWrappers;
	private ActiveComboRuleWrapper activeCombo;

	public SimpleReinforcementLearning(GameStateInterface gameState) {
		super(gameState);
		lastUpdateSystemMillis = 0;
		sumWeights = 0;
		updateScript = false;
		enemyOldHP = 0;
		myOldHP = 0;
		activeCombo = ActiveComboRuleWrapper.getSingleton();
	}

	private RuleWrapper getWeightedRandomRule(List<RuleWrapper> rules) {

		long sum = 0;

		for (RuleWrapper rule : rules) {
			sum += rule.weight;
		}

		if (sum > 0) {
			float rand = (float) (Math.random() * sum);
			float upperLimit = 0;
			for (RuleWrapper rule : rules) {
				upperLimit += rule.weight;
				if (rand <= upperLimit) {
					return rule;
				}
			}
		}

		return null;
	}

	@Override
	public Rule[] getDynamicScript(int length) {

		if (length > ruleWrapperBase.length) {
			throw new IllegalArgumentException(
					"Dynamic Script can not be longer than Rule Base");
		}

		if ((script == null) || (script.length != length)) {
			script = new RuleWrapper[length];
			updateDynamicScript(length);
			updateScript = false;
		}
		if (updateScript) {
			updateDynamicScript(length);
			updateScript = false;
		}

		return script;
	}

	private void updateDynamicScript(int length) {

		// reset script membership
		for (RuleWrapper w : ruleWrapperBase) {
			w.inScript = false;
		}

		// set default rules
		for (int i = 0; i < defaultRuleWrappers.length; i++) {
			script[i] = defaultRuleWrappers[i];
		}

		ArrayList<RuleWrapper> remainingRules = new ArrayList<>();
		remainingRules.addAll(Arrays.asList(ruleWrapperBase));
		int start = defaultRuleWrappers.length;

		// set fixed combo if available
		if (activeCombo.rule != null) {
			script[start] = activeCombo.rule;
			activeCombo.rule.inScript = true;
			remainingRules.remove(activeCombo.rule);
			start++;
		}

		// add adaptive rules
		for (int i = start; i < length; i++) {
			script[i] = getWeightedRandomRule(remainingRules);
			remainingRules.remove(script[i]);
			script[i].inScript = true;
		}

		// sort by priority
		Arrays.sort(script);
		logScript(script);
	}

	private void logScript(RuleWrapper[] script) {
		StringBuilder message = new StringBuilder();
		message.append("Dynamic Script:\n");
		for (RuleWrapper ruleWrapper : script) {
			message.append(ruleWrapper.rule.getClass().getSimpleName());
			message.append(" - ");
			message.append(ruleWrapper.weight);
			message.append(";");
			message.append(ruleWrapper.execCount);
			message.append('\n');
		}
	}

	@Override
	public void setRuleBase(Rule[] ruleBase) {
		super.setRuleBase(ruleBase);
		ruleWrapperBase = new RuleWrapper[ruleBase.length];
		sumWeights = 0;
		for (int i = 0; i < ruleBase.length; i++) {
			ruleWrapperBase[i] = new RuleWrapper(ruleBase[i]);
			// uniform distribution
			ruleWrapperBase[i].weight = INITIAL_WEIGHT;
			sumWeights += ruleWrapperBase[i].weight;
		}
		logRuleBase();
	}

	@Override
	public void setStaticRules(Rule[] defaultRules) {
		super.setStaticRules(defaultRules);
		defaultRuleWrappers = new RuleWrapper[defaultRules.length];
		for (int i = 0; i < defaultRules.length; i++) {
			defaultRuleWrappers[i] = new RuleWrapper(defaultRules[i]);
		}
	}

	@Override
	public void update() {

		// update frequency
		long currentMillis = System.currentTimeMillis();

		if (lastUpdateSystemMillis == 0) {
			lastUpdateSystemMillis = currentMillis;
		}

		if ((currentMillis - lastUpdateSystemMillis) >= UPDATE_FREQUENCY_MILLIS) {

			lastUpdateSystemMillis = currentMillis;

			updateRuleBase();

		}

	}

	private void updateRuleBase() {

		setActiveFlags();

		float fitness = calcFitness();
		long adjustment = calcAdjustment(fitness);
		long inactiveAdjustment = adjustment / 5;

		for (RuleWrapper ruleWrapper : script) {

			if (!ruleWrapper.active) {
				if ((ruleWrapper.weight > INITIAL_WEIGHT)
						|| (inactiveAdjustment > 0)
						|| (ruleWrapper.rule instanceof EmptyRule)) {
					// weight can not go below initial weight unless active or
					// empty rule
					ruleWrapper.weight += inactiveAdjustment;
				}
			} else {
				ruleWrapper.weight += adjustment;
			}

			clipWeight(ruleWrapper);

		}

		distributeRemainder();

		logRuleBase();

		resetExecCounts();

		// script needs to be updated after weight change
		updateScript = true;
	}

	private void setActiveFlags() {
		int sumExecs = 0;
		for (RuleWrapper ruleWrapper : script) {
			sumExecs += ruleWrapper.execCount;
		}
		// 10 percent threshold
		int activeThreshold = sumExecs / 10;
		for (RuleWrapper ruleWrapper : script) {
			if (ruleWrapper.execCount > activeThreshold) {
				ruleWrapper.active = true;
			}
		}
	}

	private void logRuleBase() {
		StringBuilder message = new StringBuilder();
		message.append("RuleBase:\n");
		for (RuleWrapper ruleWrapper : ruleWrapperBase) {
			message.append(ruleWrapper.rule.getClass().getSimpleName());
			message.append(" - ");
			message.append(ruleWrapper.weight);
			message.append(";");
			message.append(ruleWrapper.execCount);
			message.append('\n');
		}
	}

	private void distributeRemainder() {
		long sum = 0;

		for (RuleWrapper ruleWrapper : ruleWrapperBase) {
			sum += ruleWrapper.weight;
		}

		long remainder = (sumWeights - sum)
				/ (ruleWrapperBase.length - script.length);

		// remainder will only be distributed to rules that were not in script
		// during last update period
		for (RuleWrapper ruleWrapper : ruleWrapperBase) {
			if (!ruleWrapper.inScript) {
				if (((ruleWrapper.weight + remainder) <= W_MAX_LOW)
						&& (((ruleWrapper.weight + remainder) >= INITIAL_WEIGHT)
								|| (remainder > 0) || (ruleWrapper.rule instanceof EmptyRule))
						&& ((ruleWrapper.weight + remainder) >= W_MIN)) {
					// weight can not go below initial weight unless active or
					// empty rule
					ruleWrapper.weight += remainder;
				}
			}
		}
	}

	private void clipWeight(RuleWrapper ruleWrapper) {
		if (ruleWrapper.weight < W_MIN) {
			ruleWrapper.weight = W_MIN;
		} else if (ruleWrapper.weight > W_MAX_HIGH) {
			ruleWrapper.weight = W_MAX_HIGH;
		} else if (ruleWrapper.weight > W_MAX_LOW) {
			if ((activeCombo.rule == null) && (ruleWrapper.active)
					&& ruleWrapper.rule.isPermanentCombo) {
				activeCombo.rule = ruleWrapper;
			} else if (!ruleWrapper.equals(activeCombo.rule)) {
				ruleWrapper.weight = W_MAX_LOW;
			}
		} else {
			// [W_MIN,W_MAX_LOW]
			if (ruleWrapper.equals(activeCombo.rule)) {
				// looses active combo state
				activeCombo.rule = null;
			}
		}
	}

	private long calcAdjustment(float fitness) {
		long adjustment;
		if (fitness < BREAK_EVEN) {
			adjustment = Math.round(-P_MAX * (BREAK_EVEN - fitness)
					/ BREAK_EVEN);
		} else {
			adjustment = Math.round(R_MAX * (fitness - BREAK_EVEN)
					/ (1 - BREAK_EVEN));
		}
		return adjustment;
	}

	private void resetExecCounts() {
		for (RuleWrapper rw : ruleWrapperBase) {
			rw.execCount = 0;
			rw.active = false;
		}
	}

	private float calcFitness() {

		int enemyHP = -gameState.getEnemyCharacterData().getHp();
		int myHP = -gameState.getMyCharacterData().getHp();
		int enemyHPNow = enemyHP - enemyOldHP;
		int myHPNow = myHP - myOldHP;

		enemyOldHP = enemyHP;
		myOldHP = myHP;

		if ((enemyHPNow + myHPNow) == 0) {
			return 0.5f;
		} else {
			return ((float) enemyHPNow) / (enemyHPNow + myHPNow);
		}
	}

}
